"""
Beacon to monitor statistics from ethernet adapters

.. versionadded:: 2015.5.0
"""

import logging

import salt.utils.beacons

try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False

log = logging.getLogger(__name__)

__virtualname__ = "network_info"

__attrs = [
    "bytes_sent",
    "bytes_recv",
    "packets_sent",
    "packets_recv",
    "errin",
    "errout",
    "dropin",
    "dropout",
]


def _to_list(obj):
    """
    Convert snetinfo object to list
    """
    ret = {}

    for attr in __attrs:
        if hasattr(obj, attr):
            ret[attr] = getattr(obj, attr)
    return ret


def __virtual__():
    if not HAS_PSUTIL:
        return (
            False,
            f"Unable to load {__virtualname__} beacon: psutil library not installed",
        )
    return __virtualname__


def validate(config):
    """
    Validate the beacon configuration
    """

    VALID_ITEMS = [
        "type",
        "bytes_sent",
        "bytes_recv",
        "packets_sent",
        "packets_recv",
        "errin",
        "errout",
        "dropin",
        "dropout",
    ]

    # Configuration for load beacon should be a list of dicts
    if not isinstance(config, list):
        return False, "Configuration for network_info beacon must be a list."
    else:

        config = salt.utils.beacons.list_to_dict(config)

        for item in config.get("interfaces", {}):
            if not isinstance(config["interfaces"][item], dict):
                return (
                    False,
                    "Configuration for network_info beacon must "
                    "be a list of dictionaries.",
                )
            else:
                if not any(j in VALID_ITEMS for j in config["interfaces"][item]):
                    return (
                        False,
                        "Invalid configuration item in Beacon configuration.",
                    )
    return True, "Valid beacon configuration"


def beacon(config):
    """
    Emit the network statistics of this host.

    Specify thresholds for each network stat
    and only emit a beacon if any of them are
    exceeded.

    Emit beacon when any values are equal to
    configured values.

    .. code-block:: yaml

        beacons:
          network_info:
            - interfaces:
                eth0:
                  type: equal
                  bytes_sent: 100000
                  bytes_recv: 100000
                  packets_sent: 100000
                  packets_recv: 100000
                  errin: 100
                  errout: 100
                  dropin: 100
                  dropout: 100

    Emit beacon when any values are greater
    than configured values.

    .. code-block:: yaml

        beacons:
          network_info:
            - interfaces:
                eth0:
                  type: greater
                  bytes_sent: 100000
                  bytes_recv: 100000
                  packets_sent: 100000
                  packets_recv: 100000
                  errin: 100
                  errout: 100
                  dropin: 100
                  dropout: 100


    """
    ret = []

    config = salt.utils.beacons.list_to_dict(config)

    log.debug("psutil.net_io_counters %s", psutil.net_io_counters)

    _stats = psutil.net_io_counters(pernic=True)

    log.debug("_stats %s", _stats)
    for interface in config.get("interfaces", {}):
        if interface in _stats:
            interface_config = config["interfaces"][interface]
            _if_stats = _stats[interface]
            _diff = False
            for attr in __attrs:
                if attr in interface_config:
                    if (
                        "type" in interface_config
                        and interface_config["type"] == "equal"
                    ):
                        if getattr(_if_stats, attr, None) == int(
                            interface_config[attr]
                        ):
                            _diff = True
                    elif (
                        "type" in interface_config
                        and interface_config["type"] == "greater"
                    ):
                        if getattr(_if_stats, attr, None) > int(interface_config[attr]):
                            _diff = True
                        else:
                            log.debug("attr %s", getattr(_if_stats, attr, None))
                    else:
                        if getattr(_if_stats, attr, None) == int(
                            interface_config[attr]
                        ):
                            _diff = True
            if _diff:
                ret.append(
                    {"interface": interface, "network_info": _to_list(_if_stats)}
                )
    return ret



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest

from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def assistive(modules):
    return modules.assistive


@pytest.fixture
def osa_script(assistive):
    osa_script_path = "/usr/bin/osascript"
    try:
        assistive.install(osa_script_path, True)
        yield osa_script_path
    except CommandExecutionError as exc:
        pytest.skip(f"Unable to install {osa_script}: {exc}")
    finally:
        osa_script_ret = assistive.installed(osa_script_path)
        if osa_script_ret:
            assistive.remove(osa_script_path)


@pytest.fixture
def install_remove_pkg_name(assistive, grains):
    smile_bundle = "com.smileonmymac.textexpander"
    try:
        yield smile_bundle
    finally:
        smile_bundle_present = assistive.installed(smile_bundle)
        if smile_bundle_present:
            assistive.remove(smile_bundle)


@pytest.mark.slow_test
def test_install_and_remove(assistive, install_remove_pkg_name, grains):
    """
    Tests installing and removing a bundled ID or command to use assistive access.
    """
    try:
        ret = assistive.install(install_remove_pkg_name)
        assert ret
    except CommandExecutionError as exc:
        if grains["osmajorrelease"] != 12:
            raise exc from None
        if "attempt to write a readonly database" not in str(exc):
            raise exc from None
        pytest.skip("Test fails on MacOS 12(attempt to write a readonly database)")
    ret = assistive.remove(install_remove_pkg_name)
    assert ret


@pytest.mark.slow_test
def test_installed(assistive, osa_script):
    """
    Tests the True and False return of assistive.installed.
    """
    # OSA script should have been installed in _setup_teardown_vars function
    ret = assistive.installed(osa_script)
    assert ret
    # Clean up install
    assistive.remove(osa_script)
    # Installed should now return False
    ret = assistive.installed(osa_script)
    assert not ret


@pytest.mark.slow_test
def test_enable(assistive, osa_script):
    """
    Tests setting the enabled status of a bundled ID or command.
    """
    # OSA script should have been installed and enabled in _setup_teardown_vars function
    # Now let's disable it, which should return True.
    ret = assistive.enable(osa_script, False)
    assert ret
    # Double check the script was disabled, as intended.
    ret = assistive.enabled(osa_script)
    assert not ret
    # Now re-enable
    ret = assistive.enable(osa_script)
    assert ret
    # Double check the script was enabled, as intended.
    ret = assistive.enabled(osa_script)
    assert ret


@pytest.mark.slow_test
def test_enabled(assistive, osa_script):
    """
    Tests if a bundled ID or command is listed in assistive access returns True.
    """
    # OSA script should have been installed in _setup_teardown_vars function, which sets
    # enabled to True by default.
    ret = assistive.enabled(osa_script)
    assert ret
    # Disable OSA Script
    assistive.enable(osa_script, False)
    # Assert against new disabled status
    ret = assistive.enabled(osa_script)
    assert not ret



"""
    Tests cases for salt.modules.tomcat
"""

import io
import urllib.request

import pytest

import salt.modules.tomcat as tomcat
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {tomcat: {}}


def test_tomcat_wget_no_bytestring():
    responses = {
        "string": io.StringIO("Best response ever\r\nAnd you know it!"),
        "bytes": io.BytesIO(b"Best response ever\r\nAnd you know it!"),
    }

    string_mock = MagicMock(return_value=responses["string"])
    bytes_mock = MagicMock(return_value=responses["bytes"])
    with patch(
        "salt.modules.tomcat._auth",
        MagicMock(
            return_value=urllib.request.build_opener(
                urllib.request.HTTPBasicAuthHandler(),
                urllib.request.HTTPDigestAuthHandler(),
            )
        ),
    ):
        with patch("urllib.request.urlopen", string_mock):
            response = tomcat._wget("tomcat.wait", url="http://localhost:8080/nofail")
            for line in response["msg"]:
                assert isinstance(line, str)

        with patch("urllib.request.urlopen", bytes_mock):
            try:
                response = tomcat._wget(
                    "tomcat.wait", url="http://localhost:8080/nofail"
                )
            except TypeError as type_error:
                if (
                    type_error.args[0]
                    == "startswith first arg must be bytes or a tuple of bytes,"
                    " not str"
                ):
                    print("Got back a byte string, should've been a string")
                else:
                    raise type_error

            for line in response["msg"]:
                assert isinstance(line, str)



import logging
import shutil
from pathlib import Path

import pytest

import salt.cache
import salt.loader
from tests.pytests.functional.cache.helpers import run_common_cache_tests

log = logging.getLogger(__name__)


@pytest.fixture
def cache(minion_opts):
    opts = minion_opts.copy()
    opts["cache"] = "localfs"
    cache = salt.cache.factory(opts)
    try:
        yield cache
    finally:
        shutil.rmtree(opts["cachedir"], ignore_errors=True)


def test_caching(subtests, cache):
    run_common_cache_tests(subtests, cache)


def test_store_is_constrained_to_cachedir(cache, tmp_path):
    cache.store(str(tmp_path), "yikes", "wat")
    assert not (tmp_path / "yikes.p").exists()
    assert (Path(cache.cachedir) / str(tmp_path)[1:]).exists()


@pytest.mark.parametrize(
    "full_path,in_key", ((False, False), (False, True), (True, False))
)
def test_fetch_is_constrained_to_cachedir(cache, tmp_path, full_path, in_key):
    fake_data = b"\xa1a"
    if full_path:
        cbank = f"{tmp_path}/yikes"
        ckey = "foo"
        data = {"foo": "wat"}
        fake_data = b"\x81\xa3foo\xa1a"
    elif in_key:
        cbank = "/"
        ckey = f"{tmp_path}/yikes"
        data = "wat"
    else:
        cbank = str(tmp_path)
        ckey = "yikes"
        data = "wat"
    (tmp_path / "yikes.p").write_bytes(fake_data)
    cache.store(str(tmp_path)[1:], "yikes", data)
    res = cache.fetch(cbank, ckey)
    assert res == "wat"


def test_updated_is_constrained_to_cachedir(cache, tmp_path):
    (tmp_path / "yikes.p").touch()
    assert not cache.updated(str(tmp_path), "yikes")


@pytest.mark.parametrize("key", (None, "yikes"))
def test_flush_is_constrained_to_cachedir(cache, tmp_path, key):
    tgt = tmp_path
    if key is not None:
        tgt = tmp_path / f"{key}.p"
        tgt.touch()
    cache.flush(str(tmp_path), key)
    assert tgt.exists()


def test_list_is_constrained_to_cachedir(cache, tmp_path):
    (tmp_path / "yikes").touch()
    res = cache.list(str(tmp_path))
    assert "yikes" not in res


@pytest.mark.parametrize("key", (None, "yikes"))
def test_contains_is_constrained_to_cachedir(cache, tmp_path, key):
    if key is not None:
        (tmp_path / f"{key}.p").touch()
    assert not cache.contains(str(tmp_path), key)



"""
unittests for yaml outputter
"""

import pytest

import salt.output.yaml_out as yaml
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {yaml: {}}


@pytest.fixture
def data():
    return {"test": "two", "example": "one"}


def test_default_output(data):
    ret = yaml.output(data)
    expect = "example: one\ntest: two\n"
    assert expect == ret


def test_negative_int_output(data):
    with patch.dict(yaml.__opts__, {"output_indent": -1}):
        ret = yaml.output(data)
        expect = "{example: one, test: two}\n"
        assert expect == ret



"""
State to manage monitoring in Zenoss.

.. versionadded:: 2016.3.0

This state module depends on the 'zenoss' Salt execution module.

Allows for setting a state of minions in Zenoss using the Zenoss API. Currently Zenoss 4.x and 5.x are supported.

.. code-block:: yaml

    enable_monitoring:
      zenoss.monitored:
        - name: web01.example.com
        - device_class: /Servers/Linux
        - collector: localhost
        - prod_state: 1000
"""

import logging

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the Zenoss execution module is available.
    """
    if "zenoss.add_device" in __salt__:
        return "zenoss"
    else:
        return False, "The zenoss execution module is not available"


def monitored(name, device_class=None, collector="localhost", prod_state=None):
    """
    Ensure a device is monitored. The 'name' given will be used for Zenoss device name and should be resolvable.

    .. code-block:: yaml

        enable_monitoring:
          zenoss.monitored:
            - name: web01.example.com
            - device_class: /Servers/Linux
            - collector: localhost
            - prod_state: 1000
    """

    ret = {}
    ret["name"] = name

    # If device is already monitored, return early
    device = __salt__["zenoss.find_device"](name)
    if device:
        ret["result"] = True
        ret["changes"] = None
        ret["comment"] = f"{name} is already monitored"

        # if prod_state is set, ensure it matches with the current state
        if prod_state is not None and device["productionState"] != prod_state:
            if __opts__["test"]:
                ret["comment"] = (
                    f"{name} is already monitored but prodState will be updated"
                )
                ret["result"] = None
            else:
                __salt__["zenoss.set_prod_state"](prod_state, name)
                ret["comment"] = (
                    f"{name} is already monitored but prodState was updated"
                )

            ret["changes"] = {
                "old": "prodState == {}".format(device["productionState"]),
                "new": f"prodState == {prod_state}",
            }
        return ret

    # Device not yet in Zenoss
    if __opts__["test"]:
        ret["comment"] = f'The state of "{name}" will be changed.'
        ret["changes"] = {"old": "monitored == False", "new": "monitored == True"}
        ret["result"] = None
        return ret

    # Add and check result
    if __salt__["zenoss.add_device"](name, device_class, collector, prod_state):
        ret["result"] = True
        ret["changes"] = {"old": "monitored == False", "new": "monitored == True"}
        ret["comment"] = f"{name} has been added to Zenoss"
    else:
        ret["result"] = False
        ret["changes"] = None
        ret["comment"] = f"Unable to add {name} to Zenoss"
    return ret
